/*
Copyright (C) 2004  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "tree.h"
#include "anal.h"
//ulong   prec, precdl;

const char *Ffuncname[]={"Fseq",
		   "Fmatrix","FmatrixL","FmatrixR",
		   "Faffect",
		   "Ffacteurmat",
		   "Fmatrixelts","Fmatrixlines",
		   "Femptyvec","Femptymat","Fmat",
		   "Flistarg",

		   "Fconst","Fsmall","Fgnil",
		   "Frefarg","Ftag",
		   "Fentry","Fentryfunc","Fdeffunc",

		   /*These nodes are generated by genblock, not by parser*/
		   "Fblock"
};

const char *opname[]={"_||_", "_&&_", "_==_", "_!=_", "_>=_", "_>_", "_<=_", "_<_", "_-_","_+_","_<<_", "_>>_", "_%_", "_\\/_", "_\\_", "_/_", "_*_","_^_","__","_--","_++","_-=_", "_+=_", "_<<=_", "_>>=_", "_%=_", "_\\/=_", "_\\=_", "_/=_", "_*=_","+_","-_","!_","_!","_'","_~","",""};


typedef GEN (*PFGEN)(ANYARG);
#define _ARGS_ argvec[0], argvec[1], argvec[2], argvec[3],\
               argvec[4], argvec[5], argvec[6], argvec[7], argvec[8]

void initoperators(void)
{
  int i;
  for (i=0;i<OPnboperator;i++)
    newentry(opname[i]);
}

void
init_compiler(void)
{
  stack_init(&s_node,sizeof(*tree),(void **)&tree);
  stack_init(&s_value,sizeof(*value),(void **)&value);
  stack_init(&s_comment,sizeof(*com),(void **)&com);
  stack_init(&s_errors,sizeof(*errors),(void **)&errors);
  newnode(Fgnil,-1,-1);  /*GNIL*/
  newnode(Fgnil,-1,-1);  /*GNOARG*/
  initoperators();
}

GEN
grandopari(int a)
{
  GEN res;
  long m=1;
  if (tree[a].f!=Fentryfunc || tree[a].x!=OPpow)
    res=evalnode(a);
  else
  {
    int y=tree[a].y;
    res=evalnode(tree[y].x);
    m=itos(evalnode(tree[y].y));
  }
  return ggrando(res,m); 
}

GEN andpari(int a, int b)
{
  if (gcmp0(evalnode(a)))
    return gzero;
  return gcmp0(evalnode(b))?gzero:gun;
}

GEN orpari(int a, int b)
{
  if (!gcmp0(evalnode(a)))
    return gun;
  return gcmp0(evalnode(b))?gzero:gun;
}

GEN gmule(entree *ep, GEN x)
{
  changevalue(ep,gmul(ep->value,x));
  return ep->value;
}

GEN gdive(entree *ep, GEN x)
{
  changevalue(ep,gdiv(ep->value,x));
  return ep->value;
}

GEN gshiftle(entree *ep, long n)
{
  changevalue(ep,gshift(ep->value,n));
  return ep->value;
}

GEN gshiftre(entree *ep, long n)
{
  changevalue(ep,gshift(ep->value,-n));
  return ep->value;
}

GEN gadde(entree *ep, GEN x)
{
  changevalue(ep,gadd(ep->value,x));
  return ep->value;
}

GEN gaddone(entree *ep)
{
  changevalue(ep,gaddgs(ep->value,1));
  return ep->value;
}

GEN gsube(entree *ep, GEN x)
{
  changevalue(ep,gsub(ep->value,x));
  return ep->value;
}

GEN gsubone(entree *ep)
{
  changevalue(ep,gsubgs(ep->value,1));
  return ep->value;
}

void localpari(GEN v)
{
  return;
}

GEN gshift_right(GEN x, long n) { return gshift(x,-n); }

extern void changevalue(entree *ep, GEN x);
extern GEN gnil;

entree *getentry(int n)
{
  int x=tree[n].x;
  char *name=(char *)value[x].val.str;
  entree *ep = is_entry(name);
  if (!ep)
    ep=fetch_named_var(name,0);
  return ep;
}

static GEN
cattovec(int n, int fnum)
{
  int x=n, i=0, nb;
  GEN stack;
  while(1)
  {
    int xx=tree[x].x;
    int xy=tree[x].y;
    if (tree[x].f!=Fentryfunc || xx!=fnum) break;
    x=tree[xy].x;
    i++;
  }
  nb=i+1;
  stack=cgetg(nb+1,t_VECSMALL);
  for(x=n;i>0;i--)
  {
    int y=tree[x].y;
    x=tree[y].x;
    stack[i+1]=tree[y].y;
  }
  stack[1]=x;
  return stack;
}

VOLATILE GEN return_data;

GEN evalfunc(int n)
{
  int i=0, j=0;
  int x=tree[n].x;
  int arg[16];
  char *name=(char *)value[x].val.str;
  int ret;
  GEN res;
  char const *p,*q;
  char c;
  PPproto mod;
  entree *ep=tree[n].e;
  void *call; 
  GEN argvec[9];
  long m;
  int nb=genlistargs(n,arg,0,16);
  int has_pointer=0;
  GEN gen[9];
  entree *init[9];
  if (!ep)
  {
    ep = is_entry(name);
    if (!ep) 
      err(talker,"unknown function %s",name);
    ep=do_alias(ep);
    tree[n].e=ep;
  }
  if (ep->valence==EpUSER)
  {
    int arg2[8];
    GEN gen[8];
    GEN res;
    int nb2=genlistargs((int)ep->args,arg2,0,8);
    for (i=0;i<nb;i++)
    {
      evalnode(arg2[i]);
      gen[i]=gclone(tree[arg2[i]].e->value);
      changevalue(tree[arg2[i]].e,evalnode(arg[i]));
    }
    for (  ;i<nb2;i++)
    {
      evalnode(arg2[i]);
      gen[i]=gclone(tree[arg2[i]].e->value);
      changevalue(tree[arg2[i]].e,gzero);
    }
    
    CATCH(returner) {
      CATCH_RELEASE();
      res=gcopy((GEN)return_data);
    } TRY {
      res=evalnode((int)ep->value);
    } ENDCATCH;
    for (i=0;i<nb2;i++)
    {
      changevalue(tree[arg2[i]].e,gen[i]);
      gunclone(gen[i]);
    }
    return res;
  }
  p=ep->code; call= ep->value;
  if (!call)
    err(impl,name);
  if      (*p == 'v') { ret = RET_VOID; p++; }
  else if (*p == 'i') { ret = RET_INT;  p++; }
  else if (*p == 'l') { ret = RET_LONG; p++; }
  else ret = RET_GEN;
  q=p; 
  while((mod=parseproto(&p,&c)))
  {
    if (j<nb && arg[j]!=GNOARG)
      mod=PPstd;
    switch(mod)
    {
    case PPstd:
      switch(c)
      {
      case 'G':
        argvec[i]=evalnode(arg[j++]);
        break;
      case 'p':
        argvec[i]=(GEN)prec;
        break;
      case 'M':
      case 'L':
        argvec[i]=(GEN)itos(evalnode(arg[j++]));
        break;
      case 'n':
        argvec[i]=(GEN)varn(evalnode(arg[j++]));
        break;
      case '&':
        {
          int a=arg[j++];
          if (tree[a].f!=Frefarg)
            err(talker,"& expected");
          evalnode(a);
          has_pointer |= (1 << i);
          init[i] = tree[a].e;
          gen[i]=tree[a].e->value;
          argvec[i]=(GEN) &gen[i];
        }
        break;
      case 'P':
        argvec[i]=(GEN)precdl;
        break; 
      case 'I':
      case 'E':
        argvec[i]=(GEN)arg[j++];
        break;
      case 'S':
      case 'V':
        argvec[i]=(GEN)getentry(arg[j++]);
        break;
      case '=':
        if (tree[arg[j]].f!=Faffect)
          err(talker,"= expected");
        argvec[i]=(GEN)getentry(tree[arg[j]].x);
        arg[j]=tree[arg[j]].y;
        break;
      case '*':
        {
          int n=nb-j;
          int k,l,l1,m;
          GEN r,g=cgetg(n+1,t_VEC);
          for(l1=0,k=1;k<=n;k++)
          {
            g[k]=(long)cattovec(arg[j+k-1],OPcat);
            l1+=lg(g[k])-1;
          }
          r=cgetg(l1+1,t_VEC);
          for(m=1,k=1;k<=n;k++)
          {
            for(l=1;l<lg(g[k]);l++)
              r[m++]=(long)evalnode(mael(g,k,l));
          }
          argvec[i]=r;
        }
        break;
      case 's':
        argvec[i]=(GEN)GENtostr(evalnode(arg[j++]));
        break;
      case 'r':
        {
          int a=arg[j++];
          if (tree[a].f==Fentry)
            argvec[i]=(GEN)value[tree[a].x].val.str;
          else
            argvec[i]=(GEN)GENtostr(evalnode(a));
        }
        break;
      default:
        err(talker,"Unknown prototype code `%c' for `%s'",c,name);
      }
      break;
    case PPdefault:
      j++;
      switch(c)
      {
      case 'G':
      case '&':
      case 'V':
      case 'r':
        argvec[i]=(GEN)NULL;
        break;
      case 'n':
        argvec[i]=(GEN)-1;
        break;
      case 'E':
      case 'I':
        argvec[i]=(GEN)GNOARG;
        break;
      default:
        err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
      }
      break;
    case PPdefaultmulti:
      j++;
      switch(c)
      {
      case 'L':
      case 'M':
        argvec[i]=(GEN)strtol(q+1,NULL,10);
        break;
      case 's':
      case 'r':
        argvec[i]=(GEN)"";
        break;
      default:
        err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
      }
      break;
    default:
      err(bugparier,"PPproto %d in gencallfunc",mod);
    }
    i++;
    q=p;
  }
  for (  ;i<9;i++) argvec[i]=NULL;
  switch (ret)
  {
  case RET_GEN: 
    res = ((PFGEN)call)(_ARGS_);
    break;

  case RET_INT:
    m = (long)((int (*)(ANYARG))call)(_ARGS_);
    res = stoi(m); break;

  case RET_LONG:
    m = ((long (*)(ANYARG))call)(_ARGS_);
    res = stoi(m); break;

  case RET_VOID:
    ((void (*)(ANYARG))call)(_ARGS_);
    res = gnil; break;
  default:
    res=gnil;
  }
  if (has_pointer)
    for(i=0;i<8;i++)
      if (has_pointer&(1<<i))
        changevalue(init[i],gen[i]);
  return res;
}

GEN 
evalfacteurmat(int n)
{ 
  int x=tree[n].x;
  int y=tree[n].y;
  int yx=tree[y].x;
  int yy=tree[y].y;
  GEN gx=evalnode(x);
  long t=typ(gx);
  if (yy==-1)
  {
    long b=itos(evalnode(yx));
    if(t==t_VECSMALL)
      return stoi(gx[b]);
    if(t==t_VEC)
      return (GEN) gx[b];
  }
  else if (t==t_MAT)
  {
    long b=itos(evalnode(yx));
    long c=itos(evalnode(yy));
    return gcoeff(gx,b,c);
  }
  err(talker,"object has no components");
  return NULL; /* NOT REACHED */ 
}

GEN evalvector(int n)
{
  int arg[STACKSZ];
  int x=tree[n].x;
  int i,nb;
  GEN ret;
  nb=1+listtostack(x,Fmatrixelts,arg+1,STACKSZ-1,"Vector too long.",n);
  ret=cgetg(nb,t_VEC); 
  for(i=1;i<nb;i++)
    ret[i]=(long)evalnode(arg[i]);
  return ret; 
}

GEN evalmatrix(int n)
{
  int line[STACKSZ];
  int arg[STACKSZ];
  int x=tree[n].x;
  int i,j,nb=0,nbline,nbcol=0;
  GEN ret;
  nbline=listtostack(x,Fmatrixlines,line,STACKSZ,"Matrix too long.",n);
  for(i=0;i<nbline;i++)
  {
    int k;
    k=listtostack(line[i],Fmatrixelts,arg+nb,STACKSZ-nb,"Matrix too long.",n);
    if (i && k!=nbcol)
      err(talker,"Matrix must be rectangular");
    nbcol=k;
    nb+=k;
  }
  ret=cgetg(nbcol+1,t_MAT);
  for(i=1;i<=nbcol;i++)
  {
    GEN v=cgetg(nbline+1,t_COL);
    for(j=1; j<=nbline; j++)
      v[j]=(long) evalnode(arg[(i-1)+(j-1)*nbcol]);
    ret[i]=(long) v;
  }
  return ret;
}

GEN evalnode(int n)
{
  int x,y;
  if (n<0)
    err(bugparier,"n<0 in evalnode");
  if (DEBUGLEVEL)
    fprintferr("%s\n",Ffuncname[tree[n].f]);
  x=tree[n].x;
  y=tree[n].y;
  switch(tree[n].f)
  {
  case Fseq:
    evalnode(x);
    return evalnode(y);
  case Fmatrix:
  case FmatrixL:
  case FmatrixR:
    break;
  case Ffacteurmat:
    return evalfacteurmat(n);
  case Faffect:
    {
      GEN y=evalnode(tree[n].y);
      changevalue(getentry(tree[n].x),y);
      return y;
    }
  case Fconst:
    switch(value[x].type)
    {
    case CSTsmall:
      return stoi(value[x].val.small);
    case CSTsmallreal:
      return dbltor(value[x].val.small*1.);
    case CSTgen: 
      return value[x].val.gen;
    case CSTdot: 
      return itor(value[x].val.gen,prec);
    case CSTstr:
      return STRtoGENstr((char *)value[x].val.str);
    case CSTquote:
      return polx[fetch_user_var((char *)value[x].val.str)];
    }
    break;
  case Fsmall:
    return stoi(x);
  case Fmatrixelts:
  case Fmatrixlines:
    break;
  case Femptyvec:
    return cgetg(1,t_VEC);
  case Femptymat:
    return cgetg(1,t_MAT);
  case Fmat:
    if (tree[x].f==Fmatrixlines)
      /*if there is at least 2 lines it is a matrix*/
      return evalmatrix(n);
    else
      /*else it is a vector*/
      return evalvector(n);
  case Flistarg:
    break;
  case Frefarg:
  case Fentry:
    {
      entree *ep=tree[n].e;
      if (!ep)
      {
        ep = getentry(n);
        tree[n].e=ep;
        if (EpVALENCE(ep)!=EpVAR && EpVALENCE(ep)!=EpGVAR)
        {
          tree[n].f=Fentryfunc;
          return evalfunc(n);
        }
      }
      return ep->value;
    }
  case Fentryfunc:
    return evalfunc(n);
  case Fdeffunc:
    {
    entree *ep=getentry(x);
    ep->valence = EpUSER;
    ep->value = (void *) y;
    ep->args = (void *) x;
    return gnil;
    }
  case Fblock:
    break;
  case Fgnil:
    return gnil;
  case Ftag:
    return evalnode(x);    
  }
  err(talker,"node %s not implemented yet\n",Ffuncname[tree[n].f]);
  return gnil;
}

GEN
eval(void)
{
  return evalnode (s_node.n-1);
}
